import hashlib
import json
import os
from typing import Any, Dict, Optional
import zipfile

import requests
from seaplane.gen.carrier import ApiException, CarrierStreamOptions, FlowApi, StreamApi
from seaplane.gen.carrier.models.flow import Flow

from seaplane.config import config
from seaplane.errors import SeaplaneError
from seaplane.kv import KeyValueStorageAPI
from seaplane.logs import log
from seaplane.object import ObjectStorageAPI
from seaplane.sdk_internal_utils.http import headers
from seaplane.sdk_internal_utils.retry import retry
from seaplane.sdk_internal_utils.token_auth import get_pdk_client, with_token


def _file_md5(path: str) -> str:
    """
    Gets the MD5 hash of a file by path.
    """

    hasher = hashlib.md5()
    block_size = 4194304  # 4 MB
    with open(path, "rb") as fh:
        while True:
            buffer = fh.read(block_size)
            if not buffer:
                break
            hasher.update(buffer)
    return hasher.hexdigest()


@with_token
def create_stream(token: str, name: str) -> None:
    log.logger.debug(f"creating stream: {name}")
    stream_options: Dict[str, Any] = {}
    if config.region is not None:
        stream_options["allow_locations"] = [f"iata/{config.region}"]
    carrier_stream_options = CarrierStreamOptions.from_dict(stream_options)
    api = StreamApi(get_pdk_client(token))  # type: ignore

    def op() -> None:
        api.create_stream(
            stream_name=name,
            carrier_stream_options=carrier_stream_options,
        )

    def verify() -> Any:
        return api.get_stream(stream_name=name)

    try:
        retry(op, verify)
    except Exception:
        raise SeaplaneError(f"Failed creating stream: {name}")


@with_token
def delete_stream(token: str, name: str) -> None:
    log.logger.debug(f"deleting stream: {name}")
    api = StreamApi(get_pdk_client(token))  # type: ignore

    def op() -> None:
        try:
            api.delete_stream_with_http_info(stream_name=name)
        except ApiException as e:
            if e.status == 404:
                return None

    def verify() -> Any:
        try:
            api.get_stream_with_http_info(stream_name=name)
            return False
        except ApiException as e:
            return e.status == 404

    try:
        retry(op, verify)
    except Exception:
        raise SeaplaneError(f"Failed deleting stream: {name}")


@with_token
def add_secrets(token: str, name: str, secrets: Dict[str, str]) -> None:
    log.logger.debug(f"adding secrets: {name}")
    url = f"{config.carrier_endpoint}/flow/{name}/secrets"
    if config.region is not None:
        url += f"?region={config.region}"

    flow_secrets = {}
    for secret_key, secret_value in secrets.items():
        flow_secrets[secret_key] = {"destination": "all", "value": secret_value}

    resp = requests.put(
        url,
        json=flow_secrets,
        headers=headers(token),
    )
    resp.raise_for_status()


@with_token
def create_flow(token: str, name: str, workload: Dict[str, Any]) -> None:
    log.logger.debug(f"creating flow: {name}")
    api = FlowApi(get_pdk_client(token))  # type: ignore
    flow = Flow.from_dict(workload)

    def op() -> None:
        if config.region is not None:
            api.create_flow(
                flow_name=name,
                region=config.region,
                flow=flow,
            )
        else:
            api.create_flow(
                flow_name=name,
                flow=flow,
            )

    def verify() -> Any:
        if config.region is not None:
            return api.get_flow(flow_name=name, region=config.region)
        else:
            return api.get_flow(flow_name=name)

    try:
        retry(op, verify)
    except Exception:
        raise SeaplaneError(f"Failed creating flow: {name}")


@with_token
def delete_flow(token: str, name: str) -> None:
    log.debug(f"deleting flow: {name}")
    api = FlowApi(get_pdk_client(token))  # type: ignore

    def op() -> None:
        try:
            if config.region is not None:
                api.delete_flow_with_http_info(flow_name=name, region=config.region)
            else:
                api.delete_flow_with_http_info(flow_name=name)
        except ApiException as e:
            if e.status == 404:
                return None

    def verify() -> Any:
        try:
            if config.region is not None:
                api.get_flow_with_http_info(flow_name=name, region=config.region)
            else:
                api.get_flow_with_http_info(flow_name=name)
            return False
        except ApiException as e:
            return e.status == 404

    try:
        retry(op, verify)
    except Exception:
        raise SeaplaneError(f"Failed deleting flow: {name}")


# Note: These two functionts use KeyValueStorageAPI() for better exception handling.
def create_store(name: str, store_config: Optional[Dict[str, Any]] = None) -> None:
    """
    Creates a KV store.
    """
    log.logger.debug(f"creating store: {name}")
    if store_config is None:
        store_config = {}
    if config.region is not None:
        store_config["allow_locations"] = [f"iata/{config.region}"]
    kv = KeyValueStorageAPI()
    kv.create_store(name, body=store_config)


def delete_store(name: str) -> None:
    """
    Deletes a KV store.
    """
    log.logger.debug(f"deleting store: {name}")
    kv = KeyValueStorageAPI()
    try:
        kv.delete_store(name)
    except ApiException as e:
        if e.status != 404:
            raise


def has_store(store_name: str) -> bool:
    """
    Lists all stores in KV
    """
    kv = KeyValueStorageAPI()
    log.logger.debug("listing stores")
    store_names = kv.list_stores()
    return store_name in store_names


def set_key(store_name: str, key: str, value: str) -> None:
    """
    Set a key value in a given store
    """
    kv = KeyValueStorageAPI()
    log.logger.debug(
        f"setting key {key} in store {store_name} with value\n{json.dumps(value, indent=2)}"
    )
    kv.set_key(store_name, key, value.encode("utf-8"))


def zip_current_directory(directory_name: str) -> str:
    current_directory = os.getcwd()
    zip_filename = os.path.join(current_directory, "build", "project_for_deploy.zip")

    with zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipf.write("pyproject.toml", os.path.relpath("pyproject.toml", current_directory))

        env_file = os.environ.get("SEAPLANE_ENV_FILE", ".env")
        if os.path.exists(env_file):
            zipf.write(env_file, os.path.relpath(".env", current_directory))

        source_root = os.path.join(current_directory, directory_name)
        log.logger.debug(f"zipping directory {source_root}")
        for root, _, files in os.walk(source_root):
            for file in files:
                if "__pycache__" in root:
                    continue

                file_path = os.path.join(root, file)
                log.logger.debug(f"- zipping {file_path}")
                zipf.write(file_path, os.path.relpath(file_path, current_directory))

    return zip_filename


def upload_project(project_directory_name: str) -> str:
    """
    Zips the project directory and pushes it into the Seaplane object store,
    returning a URL that our executor image can use to refer back to the
    project when executing.

    project_directory_name should be the basename (no slashes) of the project
    directory, a peer to the pyproject.toml file for this Seaplane application.
    """
    # Step 1: Make sure we have a bucket to dump our project into
    default_bucket_name: str = "seaplane-internal-flows"
    default_bucket_config = {
        "description": "Seaplane bucket used for flow images. Should not be modified directly.",
        "replicas": 3,
        "max_bytes": -1,  # unlimited
    }

    obj = ObjectStorageAPI()
    obj._allow_internal = True
    if default_bucket_name not in obj.list_buckets():
        obj.create_bucket(default_bucket_name, default_bucket_config)

    # Step 2: Build the zip file
    project_file = zip_current_directory(project_directory_name)
    remote_path = project_directory_name + "." + _file_md5(project_file) + ".zip"

    # Step 3: Upload & return
    obj.upload_file(default_bucket_name, remote_path, project_file)

    obj_url = obj.file_url(default_bucket_name, remote_path)
    log.info(f"uploaded project {project_directory_name} => {obj_url}")

    return obj_url



from enum import Enum
from typing import Any, Dict, List, Optional
import uuid

from qdrant_client import QdrantClient
from qdrant_client.conversions import common_types as types
from qdrant_client.models import Distance as QdrantDist
from qdrant_client.models import PointStruct, PointVectors, VectorParams

from seaplane.config import config


class Vector:
    def __init__(
        self,
        vector: List[float],
        id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.vector = vector
        self.metadata = metadata
        if id is None:
            self.id = str(uuid.uuid4())
        else:
            self.id = id

    def __repr__(self) -> str:
        return f"Vector(vector={self.vector}, id={self.id}, metadata={self.metadata})"


Vectors = List[Vector]


class Distance(Enum):
    COSINE = "cosine"
    DOT = "dot"
    EUCLID = "euclid"

    def __str__(self) -> str:
        return str(self.value)


class VectorStore:
    def __init__(self) -> None:
        self.client: Optional[QdrantClient] = None
        self.TIME_OUT_IN_SECONDS = 30
        pass

    def _get_client(self) -> Optional[QdrantClient]:
        if not self.client:
            self.connect()

        return self.client

    def _get_auth_header(self, force_renew: bool = False) -> str:
        token = config._token_api.access_token
        if not token or force_renew:
            token = config._token_api.renew_token()

        return f"Bearer {token}"

    def connect(self) -> QdrantClient:
        ret = QdrantClient(
            url=config.vector_db_endpoint, port=443, timeout=self.TIME_OUT_IN_SECONDS
        )

        def auth_middleware(request: Any, call_next: Any) -> Any:
            request.headers["Authorization"] = self._get_auth_header()
            response = call_next(request)

            if response.status_code == 401:
                request.headers["Authorization"] = self._get_auth_header(force_renew=True)
                response = call_next(request)

            return response

        http_client = ret.http.client
        http_client.add_middleware(auth_middleware)

        self.client = ret

        return ret

    def map_status(self, status: Any) -> str:
        return str(status).removeprefix("UpdateStatus.")

    def check_connection(self) -> QdrantClient:
        if not self.client:
            client = self.connect()
        else:
            client = self.client

        return client

    def create_index(
        self,
        name: str,
        vector_size: int = 768,
        distance: Distance = Distance.COSINE,
    ) -> bool:
        client = self.check_connection()

        return client.create_collection(
            collection_name=name,
            vectors_config=VectorParams(
                size=vector_size, distance=self._map_to_qdrant_distance(distance)
            ),
            timeout=self.TIME_OUT_IN_SECONDS,
            replication_factor=3,
            write_consistency_factor=2,
        )

    def recreate_index(
        self,
        name: str,
        vector_size: int = 768,
        distance: Distance = Distance.COSINE,
    ) -> bool:
        client = self.check_connection()

        return client.recreate_collection(
            collection_name=name,
            vectors_config=VectorParams(
                size=vector_size, distance=self._map_to_qdrant_distance(distance)
            ),
            timeout=self.TIME_OUT_IN_SECONDS,
            replication_factor=3,
            write_consistency_factor=2,
        )

    def delete_index(self, name: str) -> bool:
        client = self.check_connection()

        return client.delete_collection(collection_name=name, timeout=self.TIME_OUT_IN_SECONDS)

    def list_indexes(self) -> List[str]:
        client = self.check_connection()

        collections = client.get_collections()
        return [idx.name for idx in collections.collections]

    def insert(self, index_name: str, vectors: Vectors) -> Dict[str, Any]:
        client = self.check_connection()

        ids = [vector.id for vector in vectors]

        result = client.upsert(
            collection_name=index_name,
            points=[
                PointStruct(id=vector.id, vector=vector.vector, payload=vector.metadata)
                for vector in vectors
            ],
        )

        return {"id": ids, "status": self.map_status(result.status)}

    def delete(self, index_name: str, points_selector: types.PointsSelector) -> str:
        """Delete Points from an index (collection)

        Points contain both vectors and metadata.
        This is the preferred method for removing data.
        """
        client = self.check_connection()

        result = client.delete(collection_name=index_name, points_selector=points_selector)

        return self.map_status(result.status)

    def delete_vectors(self, index_name: str, id: List[str]) -> str:
        """Delete vectors from an index (collection)

        This only deletes the vector, leaving the metadata behind.
        You should probably use the delete method instead!
        """
        client = self.check_connection()

        result = client.delete_vectors(collection_name=index_name, vectors=[""], points=id)

        return self.map_status(result.status)

    def update(self, index_name: str, vectors: Vectors) -> Dict[str, Any]:
        client = self.check_connection()

        ids = [vector.id for vector in vectors]

        result = client.update_vectors(
            collection_name=index_name,
            vectors=[PointVectors(id=vector.id, vector=vector.vector) for vector in vectors],
        )

        return {"id": ids, "status": self.map_status(result.status)}

    def _map_to_qdrant_distance(self, distance: Distance) -> QdrantDist:
        if distance == Distance.COSINE:
            return QdrantDist.COSINE
        elif distance == Distance.EUCLID:
            return QdrantDist.EUCLID
        elif distance == Distance.DOT:
            return QdrantDist.DOT
        else:
            return QdrantDist.COSINE

    def knn_search(
        self,
        index_name: str,
        vector: Vector,
        neighbors: int = 10,
        query_filter: Optional[types.Filter] = None,
    ) -> object:
        client = self.check_connection()

        if query_filter is not None:
            return client.search(
                collection_name=index_name,
                query_vector=vector.vector,
                limit=neighbors,
                query_filter=query_filter,
            )
        return client.search(
            collection_name=index_name, query_vector=vector.vector, limit=neighbors
        )


vector_store = VectorStore()



from importlib.metadata import version
import os
from typing import Optional

import seaplane.gen.carrier
import seaplane.config

from seaplane.logs import log
import seaplane.run_load_dotenv  # noqa: F401
from seaplane.sdk_internal_utils.token_api import TokenAPI

_SEAPLANE_IDENTIFY_API_ENDPOINT = "https://flightdeck.cplane.cloud/v1"
_SEAPLANE_GLOBAL_SQL_API_ENDPOINT = "https://sql.cplane.cloud/v1"
_SEAPLANE_SQL_DATABASE_ENDPOINT = "sql.cplane.cloud"
_SEAPLANE_CARRIER_API_ENDPOINT = "https://carrier.cplane.cloud/v1"
_SEAPLANE_SUBSTATION_EMBED_API_ENDPOINT = "https://embed.substation.cplane.cloud/v1/embed"
_SEAPLANE_VECTOR_DB_API_ENDPOINT = "https://vector-new.cplane.cloud"


def _default_runner_image() -> str:
    sdk_version = version("seaplane")
    return f"us-central1-docker.pkg.dev/artifacts-356722/sdk-apps/apps-executor:{sdk_version}"


class Configuration:
    """
    Seaplane SDK Configuration.

    Everytime the configuration is changed,
    It'll clear local configurations to the default Auth module.
    """

    def __init__(self) -> None:
        self.region: Optional[str] = None
        self.identify_endpoint = _SEAPLANE_IDENTIFY_API_ENDPOINT
        self.global_sql_endpoint = _SEAPLANE_GLOBAL_SQL_API_ENDPOINT
        self.sql_database_endpoint = _SEAPLANE_SQL_DATABASE_ENDPOINT
        self.carrier_endpoint = _SEAPLANE_CARRIER_API_ENDPOINT
        self.vector_db_endpoint = _SEAPLANE_VECTOR_DB_API_ENDPOINT
        self.substation_embed_endpoint = _SEAPLANE_SUBSTATION_EMBED_API_ENDPOINT
        self.runner_image = os.environ.get("SEAPLANE_RUNNER_IMAGE", _default_runner_image())
        self.name_prefix = os.environ.get("SEAPLANE_NAME_PREFIX", "")

        api_key = os.environ.get("SEAPLANE_API_KEY")
        self._token_api = TokenAPI(url=f"{self.identify_endpoint}/token", api_key=api_key)

    def set_api_key(self, api_key: str) -> None:
        """Set the Seaplane API Key.

        The API Key is needed for the Seaplane Python SDK usage.

        Parameters
        ----------
        api_key : str
            Seaplane API Key.
        """
        self._token_api.set_api_key(api_key)

    def get_api_key(self) -> str:
        ret = self._token_api.api_key
        if ret is None:
            raise RuntimeError(
                "missing Seaplane API key. Set SEAPLANE_API_KEY in your environment"
            )

        return ret

    def log_level(self, level: int) -> None:
        """Change logging level.
        Seaplane uses Python logging module for internal logs.
        Python logging levels can be used directly with Seaplane Python SDK or
        use the already defined in seaplane.log module.
            $ from seaplane import sea, log
            $ sea.config.log_level(log.INFO)
        Parameters
        ----------
        level : int
            Logging Level from Python logging module,
            like DEBUG, INFO, WARNING, ERROR, CRITICAL
        """
        log.level(level)

        if level == log.DEBUG:
            log.debug("Seaplane debug activated")
            log.debug(f"Identify endpoint: {self.identify_endpoint}")
            log.debug(f"Carrier endpoint: {self.carrier_endpoint}")
            log.debug(f"Vector endpoint: {self.vector_db_endpoint}")

    def staging_mode(self) -> None:
        """
        This function is for internal use by Seaplane, only.
        In general, using this will cause your applications to fail
        to deploy and run.
        """
        self.global_sql_endpoint = "https://sql.staging.cplane.dev/v1"
        self.sql_database_endpoint = "sql.staging.cplane.dev"
        self.carrier_endpoint = "https://carrier.staging.cplane.dev/v1"
        self.identify_endpoint = "https://flightdeck.staging.cplane.dev/v1"
        self.vector_db_endpoint = "https://vector-new.staging.cplane.dev"
        self._token_api.set_url(url=f"{self.identify_endpoint}/token")

    def set_region(self, region: str) -> None:
        self.region = region.lower()

    def get_platform_configuration(self) -> seaplane.gen.carrier.Configuration:
        configuration = seaplane.gen.carrier.Configuration()  # type: ignore
        configuration.host = self.carrier_endpoint
        return configuration

    def get_access_token(self) -> str | None:
        return self._token_api.access_token


config = Configuration()



import logging
import os
import traceback
from typing import List

import seaplane.run_load_dotenv as denv


def _format_exception(e: Exception) -> List[str]:
    return traceback.format_exception(type(e), e, e.__traceback__)


class SeaLogger:
    """
    Seaplane Logger container module, which uses Python logger.
    """

    PREFIX = "[Seaplane] "
    FORMAT = "[Seaplane] %(message)s"

    CRITICAL = logging.CRITICAL
    FATAL = logging.FATAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    WARN = logging.WARN
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET

    def __init__(self) -> None:
        logging.basicConfig(format=self.FORMAT)
        self.logger = logging.getLogger("Seaplane")

        level = os.environ.get("SEAPLANE_LOGLEVEL", "INFO")
        self.logger.setLevel(level=level)

        self.logger.debug(f"loaded environment {denv.root_env} ({denv.loaded})")

    def debug(self, message: str) -> None:
        self.logger.debug(message)

    def info(self, message: str) -> None:
        self.logger.info(message)

    def error(self, message: str) -> None:
        self.logger.error(message)

    def level(self, level: int) -> None:
        """Change logging level.

        Seaplane uses Python logging module for internal logs.
        Python logging levels can be used directly with Seaplane Python SDK or
        use the already defined in seaplane.log module.

        Parameters
        ----------
        level : int
            Logging Level from Python logging module,
            like DEBUG, INFO, WARNING, ERROR, CRITICAL
        """
        self.logger.setLevel(level)

    def disable(self) -> None:
        """Disable the Seaplane logging for the SDK."""
        self.logger.setLevel(self.CRITICAL)

    def enable(self, level: int = WARNING) -> None:
        """Enable the Seaplane logging for the SDK,
        indicating the log level if needed.

        Parameters
        ----------
        level : int
            Logging Level from Python logging module,
            like DEBUG, INFO, WARNING, ERROR, CRITICAL
        """
        self.logger.setLevel(level)


log = SeaLogger()



from typing import Any, Dict, List, Optional, TypedDict, cast

import requests
from seaplane.gen.carrier import ApiException, Bucket, ObjectApi

from seaplane.config import config
from seaplane.errors import HTTPError, SeaplaneError
from seaplane.sdk_internal_utils.http import headers
from seaplane.sdk_internal_utils.retry import retry
from seaplane.sdk_internal_utils.token_auth import get_pdk_client, method_with_token

SP_BUCKETS = ["seaplane-internal-flows"]


class ListObjectsMetadata(TypedDict):
    """
    Dictionary wrapping the metadata returned by the `list_objects()` endpoint.
    """

    name: str
    digest: str
    mod_time: int
    size: int


class ObjectStorageAPI:
    """
    Class for handling Object Storage API calls.
    """

    _allow_internal: bool
    """
    If set, allows the wrapper to manipulate Seaplane-internal buckets.

    Should not be set in customer code!
    """

    def __init__(self) -> None:
        self._allow_internal = False

    def get_object_api(self, access_token: str) -> ObjectApi:
        return ObjectApi(get_pdk_client(access_token))  # type: ignore

    @method_with_token
    def list_buckets(self, token: str) -> List[str]:
        """
        Returns a list of buckets
        """
        api = self.get_object_api(token)
        list = []
        resp = api.list_buckets()
        for name, _ in sorted(resp.items()):
            if self._allow_internal or name not in SP_BUCKETS:
                list.append(name)

        return list

    @method_with_token
    def get_bucket(self, token: str, bucket_name: str) -> Dict[str, Any]:
        """
        Get metadata about a bucket
        """
        api = self.get_object_api(token)
        resp = api.get_bucket(bucket_name=bucket_name)
        return {k: cast(str, v) for (k, v) in resp.to_dict().items()}  # type: ignore

    @method_with_token
    def create_bucket(
        self, token: str, bucket_name: str, body: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Create a new bucket

        Optional body argument can be used to configure the bucket with description, notify, etc.
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot create bucket with Seaplane-internal name `{bucket_name}`"
            )

        if not body:
            body = {}
        bucket_config = Bucket.from_dict(body)
        api = self.get_object_api(token)

        def op() -> None:
            return api.create_bucket(
                bucket_name=bucket_name,
                bucket=bucket_config,
            )

        def verify() -> Bucket:
            return api.get_bucket(
                bucket_name=bucket_name,
            )

        try:
            retry(op, verify)
        except Exception:
            raise SeaplaneError(f"Failed creating bucket: {bucket_name}")

    @method_with_token
    def delete_bucket(self, token: str, bucket_name: str) -> None:
        """
        Delete a bucket
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot delete bucket with Seaplane-internal name `{bucket_name}`"
            )
        api = self.get_object_api(token)

        def op() -> None:
            try:
                api.delete_bucket_with_http_info(bucket_name=bucket_name)
            except ApiException as e:
                if e.status == 404:
                    return None

        def verify() -> Any:
            try:
                api.get_bucket_with_http_info(bucket_name=bucket_name)
                return False
            except ApiException as e:
                return e.status == 404

        try:
            retry(op, verify)
        except Exception:
            raise SeaplaneError(f"Failed deleting bucket: {bucket_name}")

    def exists(self, bucket: str, object: str) -> bool:
        """
        Returns True if object exists in bucket
        """
        try:
            objs = self.list_objects(bucket, "")
            for obj in objs:
                if obj["name"] == object:
                    return True
            return False
        except HTTPError as e:
            if e.status == 404:
                return False
            raise e

    @method_with_token
    def list_objects(
        self, token: str, bucket_name: str, path_prefix: str = "/"
    ) -> List[ListObjectsMetadata]:
        """
        List objects in a bucket, (optional) matching a path prefix

        Returns a list of dicts with name, digest, mod_time, and bytes for each object
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot list objects in bucket with Seaplane-internal name `{bucket_name}`"
            )
        api = self.get_object_api(token)

        resp = api.list_objects(
            bucket_name=bucket_name,
            path=path_prefix,
        )

        table = [
            ListObjectsMetadata(
                name=x.name,
                digest=x.digest,
                mod_time=x.mod_time,
                size=x.size,
            )
            for x in resp
        ]

        return table

    @method_with_token
    def download(self, token: str, bucket_name: str, path: str) -> bytes:
        """
        Downloads an object

        Returns the object in bytes
        """
        url = f"{config.carrier_endpoint}/object/{bucket_name}/store"

        params: Dict[str, Any] = {}
        params["path"] = path
        resp = requests.get(
            url,
            params=params,
            headers=headers(token, "application/octet-stream"),
        )
        return resp.content

    def file_url(self, bucket_name: str, path: str) -> str:
        """
        Builds a URL usable to download the object stored at the given bucket & path.
        """
        return f"{config.carrier_endpoint}/object/{bucket_name}/store?path={path}"

    @method_with_token
    def upload(self, token: str, bucket_name: str, object_path: str, object: bytes) -> None:
        """
        Create a new object with name `object_path` from the given object data (bytes)
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot upload to bucket with Seaplane-internal name `{bucket_name}`"
            )
        api = self.get_object_api(token)

        # Timeouts are expected here. Wait, confirm failure, then retry.
        # A timeout doesn't necessarily mean the operation failed
        def op() -> None:
            return api.create_object(
                bucket_name=bucket_name,
                path=object_path,
                body=object,
            )

        def verify() -> bytearray:
            return api.get_object(
                bucket_name=bucket_name,
                path=object_path,
            )

        try:
            retry(op, verify)
        except Exception:
            raise SeaplaneError(f"Failed uploading file: {object_path}")

    def upload_file(self, bucket_name: str, object_path: str, file_path: str) -> None:
        """
        Upload a local file (`file_path`) to a new object (`object_path`)
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot upload to bucket with Seaplane-internal name `{bucket_name}`"
            )
        with open(file_path, "rb") as file:
            file_data = file.read()

        self.upload(bucket_name, object_path, file_data)

    @method_with_token
    def delete(self, token: str, bucket_name: str, path: str) -> None:
        """
        Delete an object
        """
        if not self._allow_internal and bucket_name in SP_BUCKETS:
            raise SeaplaneError(
                f"Cannot delete from bucket with Seaplane-internal name `{bucket_name}`"
            )
        api = self.get_object_api(token)

        def op() -> None:
            try:
                api.delete_object_with_http_info(
                    bucket_name=bucket_name,
                    path=path,
                )
            except ApiException as e:
                if e.status == 404:
                    return None

        def verify() -> Any:
            try:
                api.get_object_with_http_info(
                    bucket_name=bucket_name,
                    path=path,
                )
                return False
            except ApiException as e:
                return e.status == 404

        try:
            retry(op, verify)
        except Exception:
            raise SeaplaneError(f"Failed deleting object: {path}")


object_store = ObjectStorageAPI()



import time
from typing import Any, Callable

from seaplane.gen.carrier import ApiException
from seaplane.logs import log


def retry(
    op: Callable[[], Any], verify: Callable[[], Any], max_retries: int = 3, delay: int = 10
) -> Any:
    """
    Performs a given operation, and retries it up to `max_retries` times if it fails.
    This function also takes an optional `verify` function, which is called when the
    operation times out, to check for the case where the operation actually succeeded.
    Returns the result of whichever operation finally succeeds (op or verify).
    """
    last_err = None

    for _ in range(1, max_retries + 1):
        try:
            return op()
        except ApiException as e:
            last_err = e

            # Any error except timeout
            if e.status != 500:
                raise

            log.logger.debug("retrying...")
            # Generous backoff, then attempt to verify the operation
            # in case it just took a long time but succeeded
            time.sleep(delay)

            if not verify:  # type: ignore
                continue

            try:
                return verify()
            except Exception:
                # Failure to verify, so retry the whole operation
                pass

    raise ApiException from last_err



import json
from typing import Any, Dict, List

import requests
from tabulate import tabulate

from seaplane.config import config
from seaplane.errors import SeaplaneError
from seaplane.kv import KeyValueStorageAPI
from seaplane.logs import log
from seaplane.pipes import App
from seaplane.sdk_internal_utils.http import headers
from seaplane.sdk_internal_utils.token_auth import with_token

ENDPOINTS_STREAM = "_SEAPLANE_ENDPOINT"
SUBSTATION_RESULTS_STREAM = "_SEAPLANE_AI_RESULTS"
SP_APP_STORE = "_SP_APP_STORE"


@with_token
def get_messages(token: str, subject: str) -> str:
    """
    Returns a string with the number of messages in stream subject
      OR "Stream not found" if it does not exist
      OR "0" if the stream exists but subject does not
    """
    stream_name = subject.split(".")[0]
    url = f"{config.carrier_endpoint}/stream/{stream_name}"

    resp = requests.get(
        url,
        headers=headers(token),
    )
    if resp.status_code == 404:
        return "N/A"

    messages = 0
    stream_info = json.loads(resp.content)
    if stream_info.get("details") is not None:
        for key in stream_info["details"]["subjects"].keys():
            if key[0 : len(subject)] == subject:
                messages += stream_info["details"]["subjects"][key]

    return str(messages)


@with_token
def get_status(token: str, flow_name: str) -> str:
    """
    Returns the status of a flow
      OR "N/A" if status not available
    """
    url = f"{config.carrier_endpoint}/flow/{flow_name}/status"
    if config.region is not None:
        url += f"?region={config.region}"
    resp = requests.get(
        url,
        headers=headers(token),
    )
    if resp.status_code == 200:
        status = json.loads(resp.content)
        replicas = 1
        dead = 0
        output: List[str] = []
        for alloc in status.keys():
            if status[alloc] != "dead":
                output.append(f"{replicas}: {status[alloc]}")
                # We could instead show the alloc id, but it gets busy...
                # output.append(f"{alloc}: {status[alloc]}")
                replicas += 1
            else:
                dead += 1
        if dead > 0:
            output.append(f"{dead} completed task(s)")
        return "\n".join(output)
    return "N/A"


def sort_nodes(app_rep: Dict[str, Any]) -> List[Dict[str, str]]:
    """
    Returns a list of nodes sorted by their edges from IN to OUT
    """
    nodes = app_rep["graph"]["nodes"]
    edges = app_rep["graph"]["edges"]

    in_nodes = [node for node in nodes if node["id"] == "in"]
    sorted_nodes = in_nodes[:]
    for sorted_node in sorted_nodes:
        target_edges = [edge for edge in edges if edge["source"] == sorted_node["id"]]
        for target_edge in target_edges:
            target_nodes = [node for node in nodes if node["id"] == target_edge["target"]]
            for target_node in target_nodes:
                if target_node not in sorted_nodes:
                    sorted_nodes.append(target_node)
            target_edge_source_edges = [
                edge
                for edge in edges
                if (edge["target"] == target_edge["target"] and edge != target_edge)
            ]
            for target_edge_source_edge in target_edge_source_edges:
                target_edge_source_nodes = [
                    node for node in nodes if node["id"] == target_edge_source_edge["source"]
                ]
                for target_edge_source_node in target_edge_source_nodes:
                    if (
                        target_edge_source_node.get("substation_id")
                        and "subscription" in target_edge_source_node["id"]
                        and target_edge_source_node not in sorted_nodes
                    ):
                        sorted_nodes.append(target_edge_source_node)

    return sorted_nodes


def status(app: App) -> None:
    """
    Prints the status of resources associated with an app
    """
    config.log_level(log.CRITICAL)
    kv = KeyValueStorageAPI()
    try:
        app_rep = json.loads(kv.get_key(SP_APP_STORE, app.name))
    except Exception:
        raise SeaplaneError(
            "Cannot load app representation. Try moving to the dir where you deploy your app."
        )

    table_headers = ["Task Name", "Status", "Messages In", "Messages Out"]
    table: List[Any] = []

    edges = app_rep["graph"]["edges"]
    sorted_nodes = sort_nodes(app_rep)
    for node in sorted_nodes:
        if node["id"] in ("in", "out") or (
            node.get("substation_id") and "subscription" in node["id"]
        ):
            continue
        node_name = node["name"]
        node_status = get_status(node_name)
        outputs = [edge["target"] for edge in edges if edge["source"] == node["id"]]
        inputs = [edge["source"] for edge in edges if edge["target"] == node["id"]]
        in_list: List[Any] = []
        for input in inputs:
            if input == "in":
                endpoint = f"{ENDPOINTS_STREAM}.in.{app.name}"
                in_list.insert(0, f"IN: {get_messages(endpoint)}")
                continue

            stream, subject = input.split(".")[0:2]
            if subject.split("-")[-1] == "subscription":
                full_subject = f"{SUBSTATION_RESULTS_STREAM}.{app.name}"
                in_list.append(f"SUBSTATION: {get_messages(full_subject)}")
            else:
                full_subject = f"{stream}.{subject}"
                in_list.append(f"{subject}: {get_messages(full_subject)}")
        messages_in = "\n".join(in_list)

        out_list = []
        for output in outputs:
            if output == "out":
                endpoint = f"{ENDPOINTS_STREAM}.out.{app.name}"
                out_list.append(f"OUT: {get_messages(endpoint)}")
                continue

            stream, subject = node["id"].split(".")[0:2]
            full_subject = f"{stream}.{subject}"
            out_list.insert(0, f"{subject}: {get_messages(full_subject)}")
        messages_out = "\n".join(out_list)

        table.append([node_name, node_status, messages_in, messages_out])

    print("\n")
    # "fancy_grid" handles newlines better than "outline"
    print(tabulate(table, table_headers, tablefmt="fancy_grid"))



from functools import wraps
from importlib.metadata import version
from typing import Callable, Concatenate, ParamSpec, TypeVar, Union

import requests
from seaplane.gen.carrier import ApiClient, ApiException, __version__

from seaplane.config import config
from seaplane.errors import HTTPError
from seaplane.logs import log
from seaplane.sdk_internal_utils.http import SDK_HTTP_ERROR_CODE
from seaplane.sdk_internal_utils.token_api import TokenAPI

_R = TypeVar("_R")
_P = ParamSpec("_P")
_SelfType = TypeVar("_SelfType")


# (self, token, *args, **kwargs) -> ret ==>  (self, *args, **kwargs) -> ret
#
# The type annotation is complicated because of PEP-612 only allowing us to put
# arguments at the front of wrapped functions, combined with methods always
# taking `self` as the first argument. We double-wrap using an explicit
# `_SelfType` so the token is passed as the first non-`self` parameter.
def method_with_token(
    func: Callable[Concatenate[_SelfType, str, _P], _R]
) -> Callable[Concatenate[_SelfType, _P], _R]:
    """
    A decorator around class methods that need JWTs.

    The JWT will be passed in as the first parameter.

    The JWTs will be pulled from the global `config` wrapper. The wrapped
    function will be retried if the token needs to be refreshed.
    """

    @wraps(func)
    def _inner(this: _SelfType, /, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        token_api = config._token_api
        try:
            return func(this, token_api.get_token(), *args, **kwargs)
        except ApiException as e:
            _renew_if_failed(token_api, e)
            return func(this, token_api.get_token(), *args, **kwargs)
        except HTTPError as e:
            _renew_if_failed(token_api, e)
            return func(this, token_api.get_token(), *args, **kwargs)
        except requests.exceptions.RequestException as err:
            _renew_if_failed(token_api, _map_request_exception(err))
            return func(this, token_api.get_token(), *args, **kwargs)

    return _inner


def with_token(func: Callable[Concatenate[str, _P], _R]) -> Callable[_P, _R]:
    """
    A decorator around bare functions that need JWTs.

    The JWT will be passed in as the first parameter.

    The JWTs will be pulled from the global `config` wrapper. The wrapped
    function will be retried if the token needs to be refreshed.
    """

    @wraps(func)
    def _inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
        token_api = config._token_api
        try:
            return func(token_api.get_token(), *args, **kwargs)
        except ApiException as e:
            _renew_if_failed(token_api, e)
            return func(token_api.get_token(), *args, **kwargs)
        except HTTPError as e:
            _renew_if_failed(token_api, e)
            return func(token_api.get_token(), *args, **kwargs)
        except requests.exceptions.RequestException as err:
            _renew_if_failed(token_api, _map_request_exception(err))
            return func(token_api.get_token(), *args, **kwargs)

    return _inner


def _renew_if_failed(token_api: TokenAPI, exc: Union[ApiException, HTTPError]) -> None:
    """
    Attempts to renew a token if the given error calls for it.
    """
    if not token_api.auto_renew:
        raise exc

    if isinstance(exc, ApiException) and exc.status != 401:
        raise exc

    if isinstance(exc, HTTPError) and "401" not in exc.message:
        raise exc

    log.info("Auto-Renew, renewing the token...")
    token_api.renew_token()


def _map_request_exception(err: requests.exceptions.RequestException) -> HTTPError:
    """
    Maps a raw `requests` exception into a Seaplane `HTTPError`.
    """
    log.error(f"Request exception: {str(err)}")
    status_code: int = SDK_HTTP_ERROR_CODE
    if err.response:
        status_code = err.response.status_code
    return HTTPError(status_code, str(err))


def get_pdk_client(access_token: str) -> ApiClient:
    """
    Constructs a Seaplane PDK ApiClient from the given access token.
    """

    pdk_config = config.get_platform_configuration()
    pdk_config.access_token = access_token
    client = ApiClient(pdk_config)  # type: ignore
    client.set_default_header("X-Seaplane-Sdk-Version", version("seaplane"))  # type: ignore
    client.set_default_header("X-Seaplane-Pdk-Version", __version__)  # type: ignore
    return client
